#include "testApp.h"

#import "PGMidi.h"
#import "iOSVersionDetection.h"
#include <mach/mach_time.h>

PGMidi * ofxMidiContext::midi = NULL;

//--------------------------------------------------------------
// from https://developer.apple.com/library/mac/#qa/qa1398/_index.html
uint64_t AbsoluteToNanos(uint64_t time) {
	static struct mach_timebase_info timebaseInfo;

	// only init once
	if(timebaseInfo.denom == 0) {
		mach_timebase_info(&timebaseInfo);
	}

	return time * timebaseInfo.numer / timebaseInfo.denom;
}

//--------------------------------------------------------------
@implementation ofxMidiDelegate

@synthesize midi;

- (id) init {
	self = [super init];
	lastTime = 0;
	firstPacket = false;
	return self;
}

// adapted from RTMidi CoreMidi message parsing
- (void) midiSource:(PGMidiSource*)midi midiReceived:(const MIDIPacketList *)packetList {

    const MIDIPacket *packet = &packetList->packet[0];
	stringstream msg;
	unsigned long long time;
	double delta;
	
    for(int i = 0; i < packetList->numPackets; ++i) {
       
		// calc time stamp
		time = 0;
		if(firstPacket) {
			firstPacket = false;
		}
		else {
			time = packet->timeStamp;
			if(time == 0) { // this happens when receiving asynchronous sysex messages
				time = mach_absolute_time();
			}
			time -= lastTime;
			delta = AbsoluteToNanos(time) * 0.000001; // convert to ms
		}
		lastTime = packet->timeStamp;
		if(lastTime == 0 ) { // this happens when receiving asynchronous sysex messages
		  lastTime = mach_absolute_time();
		}
	   
		msg << "packet: " << packet->length << " [ ";
		for(int b = 0; b < packet->length; ++b) {
			msg << std::hex << std::uppercase << std::setw(2) << std::setfill('0')
				<< (int) packet->data[b] << " " 
				<< std::dec << std::nouppercase << std::setw(1) << std::setfill(' ');
		}
		msg << "] " << delta;
		app->addMessage(msg.str());
		msg.str("");
		
		packet = MIDIPacketNext(packet);
    }
}

- (void) midi:(PGMidi*)midi sourceAdded:(PGMidiSource *)source {
	source.delegate = self;
	stringstream msg;
	msg << "source added: " << [source.name UTF8String];
	app->addMessage(msg.str());
}

- (void) midi:(PGMidi*)midi sourceRemoved:(PGMidiSource *)source {
	source.delegate = nil;
	stringstream msg;
	msg << "source removed: " << [source.name UTF8String];
	app->addMessage(msg.str());
}
- (void) midi:(PGMidi*)midi destinationAdded:(PGMidiDestination *)destination {
	stringstream msg;
	msg << "dest added: " << [destination.name UTF8String];
	app->addMessage(msg.str());
}

- (void) midi:(PGMidi*)midi destinationRemoved:(PGMidiDestination *)destination {
	stringstream msg;
	msg << "dest removed: " << [destination.name UTF8String];
	app->addMessage(msg.str());
}

- (void) attachToAllExistingSources {
    for(PGMidiSource * source in midi.sources) {
        source.delegate = self;
    }
}

- (void) setMidi:(PGMidi *)m {
    midi.delegate = nil;
    midi = m;
    midi.delegate = self;
    [self attachToAllExistingSources];
}

- (void) setApp:(void *)a {
	app = (testApp*) a;
}

- (void) listSources {
	cout << "Sources" << endl;
    for(PGMidiSource * source in midi.sources) {
        cout << "\t" << [source.name UTF8String] << endl;
    }
}

- (void) listDestinations {
	cout << "Destinations" << endl;
    for(PGMidiDestination * dest in midi.destinations) {
        cout << "\t" << [dest.name UTF8String] << endl;
    }
}

@end

//--------------------------------------------------------------
void testApp::setup(){	

	ofSetVerticalSync(true);

	// register touch events
	ofRegisterTouchEvents(this);
	
	// initialize the accelerometer
	ofxAccelerometer.setup();
	
	//iPhoneAlerts will be sent to this.
	ofxiPhoneAlerts.addListener(this);
	
	//If you want a landscape oreintation 
	//iPhoneSetOrientation(OFXIPHONE_ORIENTATION_LANDSCAPE_RIGHT);
	
	ofBackground(127,127,127);
	
	// setup midi
	ofxMidiContext::setup();
	ofxMidiContext::enableNetwork(true);
	midiDelegate = [[ofxMidiDelegate alloc] init];
	[midiDelegate setMidi:ofxMidiContext::getMidi()];
	[midiDelegate setApp:(void*) this];
	
	ofxMidiContext::listSources();
	ofxMidiContext::listDestinations();
	
	maxMessages = 28;
	messages.push_back("nothing yet ...");
	
	note = -1;
	ctl = -1;
}

//--------------------------------------------------------------
void testApp::update(){

}

//--------------------------------------------------------------
void testApp::draw(){

	ofSetColor(0);

	ofDrawBitmapString("Input:", 10, 20);
	int x = 10, y = 34;
	deque<string>::iterator iter = messages.begin();
	for(; iter != messages.end(); ++iter) {
		ofDrawBitmapString((*iter), x, y);
		y += 14;
	}
	
	ofDrawBitmapString("Output:", 10, ofGetHeight()-42);
	if(note > 0) {
		ofDrawBitmapString("note "+ofToString(note), 10, ofGetHeight()-28);
	}
	if(ctl > 0) {
		ofDrawBitmapString("pan "+ofToString(ctl), 10, ofGetHeight()-14);
	}
}

//--------------------------------------------------------------
void testApp::exit(){

}

//--------------------------------------------------------------
void testApp::touchDown(ofTouchEventArgs &touch){

	// send note on
	note = (int) ofMap(touch.y, 0, ofGetHeight(), 0, 127); 
	bytes.clear();
	bytes.push_back(0x90);
	bytes.push_back(note);
	bytes.push_back(0x40);
	[midiDelegate.midi sendBytes:(UInt8*) &bytes[0] size:bytes.size()];
}

//--------------------------------------------------------------
void testApp::touchMoved(ofTouchEventArgs &touch){
	
	// send ctl change
	ctl = (int) ofMap(touch.x, 0, ofGetWidth(), 0, 127); 
	bytes.clear();
	bytes.push_back(0xB0);
	bytes.push_back(10);	// pan
	bytes.push_back(ctl);
	[midiDelegate.midi sendBytes:(UInt8*) &bytes[0] size:bytes.size()];
}

//--------------------------------------------------------------
void testApp::touchUp(ofTouchEventArgs &touch){

	// send note off
	bytes.clear();
	bytes.push_back(0x80);
	bytes.push_back(note);
	bytes.push_back(0x40);
	[midiDelegate.midi sendBytes:(UInt8*) &bytes[0] size:bytes.size()];
	note = -1;
	ctl = -1;
}

//--------------------------------------------------------------
void testApp::touchDoubleTap(ofTouchEventArgs &touch){

}

//--------------------------------------------------------------
void testApp::lostFocus(){

}

//--------------------------------------------------------------
void testApp::gotFocus(){

}

//--------------------------------------------------------------
void testApp::gotMemoryWarning(){

}

//--------------------------------------------------------------
void testApp::deviceOrientationChanged(int newOrientation){

}


//--------------------------------------------------------------
void testApp::touchCancelled(ofTouchEventArgs& args){

}

//--------------------------------------------------------------
void testApp::addMessage(string msg) {
	messages.push_back(msg);
	while(messages.size() > maxMessages)
		messages.pop_front();
}

